@@ -1,7 +1,10 @@
-from vp_suite.models._base_model import VideoPredictionModel
+from vp_suite.base.base_model import VideoPredictionModel
 
 
 class CopyLastFrame(VideoPredictionModel):
+    r"""
+
+    """
 
     # model-specific constants
     NAME = "CopyLastFrame"
@@ -9,7 +12,22 @@ class CopyLastFrame(VideoPredictionModel):
     TRAINABLE = False
 
     def __init__(self, device=None, **model_args):
+        r"""
+
+        Args:
+            device ():
+            **model_args ():
+        """
         super(CopyLastFrame, self).__init__(device, **model_args)
 
     def pred_1(self, x, **kwargs):
+        r"""
+
+        Args:
+            x ():
+            **kwargs ():
+
+        Returns:
+
+        """
         return x[:, -1, :, :, :]
