@@ -1,4 +1,5 @@
 import random, json, os
+import warnings
 from typing import List, Dict, Any
 from pathlib import Path
 
@@ -10,9 +11,9 @@ import wandb
 from tqdm import tqdm
 
 import vp_suite.constants as constants
-from vp_suite.dataset._wrapper import DatasetWrapper
-from vp_suite.dataset import DATASET_CLASSES
-from vp_suite.models._base_model import VideoPredictionModel
+from vp_suite.utils.dataset_wrapper import DatasetWrapper
+from vp_suite.datasets import DATASET_CLASSES
+from vp_suite.base.base_model import VideoPredictionModel
 from vp_suite.models import MODEL_CLASSES, AVAILABLE_MODELS
 from vp_suite.models.copy_last_frame import CopyLastFrame
 from vp_suite.measure import LOSS_CLASSES
@@ -20,15 +21,29 @@ from vp_suite.measure.loss_provider import PredictionLossProvider
 from vp_suite.measure.metric_provider import PredictionMetricProvider
 from vp_suite.utils.visualization import visualize_vid
 from vp_suite.utils.utils import timestamp
-from vp_suite.utils.img_processor import ImgProcessor
 from vp_suite.utils.compatibility import check_model_and_data_compat, check_run_and_model_compat
 
 
 class VPSuite:
+    r"""TODO.
 
-    DEFAULT_RUN_CONFIG = (constants.PKG_RESOURCES / 'run_config.json').resolve()
+    TODO.
+
+    Attributes:
+        device(str): A string specifying which device to work on ('cuda' or 'gpu')
+        datasets(List[:class:`DatasetWrapper`]): The loaded datasets.
+        models(List[:class:`VideoPredictionModel`]): The loaded/created models.
+
+    """
+
+    _DEFAULT_RUN_CONFIG : Path = constants.PKG_RESOURCES / 'run_config.json'
 
     def __init__(self, device="cuda"):
+        r"""
+
+        Args:
+            device ():
+        """
         self.device = "cuda" if device == "cuda" and torch.cuda.is_available() else "cpu"
 
         self.clear_models()
@@ -36,59 +51,104 @@ class VPSuite:
 
     @property
     def training_sets(self):
+        r"""
+
+        Returns:
+
+        """
         return [d for d in self.datasets if d.is_training_set()]
 
     @property
     def test_sets(self):
+        r"""
+
+        Returns:
+
+        """
         return [d for d in self.datasets if d.is_test_set()]
 
     def clear_datasets(self):
+        r"""
+
+        Returns:
+
+        """
         self.datasets : List[DatasetWrapper] = []
 
     def clear_models(self):
+        r"""
+
+        Returns:
+
+        """
         self.models : List[VideoPredictionModel] = []
 
-    def load_dataset(self, dataset="MM", split="train", value_min=0.0, value_max=1.0, **dataset_kwargs):
+    def load_dataset(self, dataset, split="train", **dataset_kwargs):
+        r"""
+
+        Args:
+            dataset ():
+            split ():
+            value_min ():
+            value_max ():
+            **dataset_kwargs ():
+
+        Returns:
+
         """
-        ATTENTION TODO
-        """
-        img_processor = ImgProcessor(value_min=value_min, value_max=value_max)
         dataset_class = DATASET_CLASSES[dataset]
-        dataset = DatasetWrapper(dataset_class, img_processor, split, **dataset_kwargs)
-        print(f"INFO: loaded dataset '{dataset.NAME}' from {dataset.data_dir} "
+        dataset = DatasetWrapper(dataset_class, split, **dataset_kwargs)
+        print(f"loaded dataset '{dataset.NAME}' from {dataset.data_dir} "
               f"(action size: {dataset.action_size})")
         self.datasets.append(dataset)
 
-    def load_model(self, model_dir, ckpt_name="best_model.pth"):
-        """
-        overrides existing model
-        """
+    def download_dataset(self, dataset):
+        dataset_class = DATASET_CLASSES[dataset]
+        dataset_class.download_and_prepare_dataset()
 
+    def load_model(self, model_dir, ckpt_name="best_model.pth"):
+        r"""
+
+        Args:
+            model_dir ():
+            ckpt_name ():
+
+        Returns:
+
+        """
         model_ckpt = os.path.join(model_dir, ckpt_name)
         model = torch.load(model_ckpt)
         self._model_setup(model, loaded=True)
 
     def create_model(self, model_type, action_conditional=False, **model_args):
-        """
-        overrides existing model
+        r"""
+
+        Args:
+            model_type ():
+            action_conditional ():
+            **model_args ():
+
+        Returns:
+
         """
 
         # parameter processing
-        assert model_type in AVAILABLE_MODELS, f"ERROR: invalid model type specified! " \
-                                               f"Available model types: {AVAILABLE_MODELS}"
+        if model_type not in AVAILABLE_MODELS:
+            raise ValueError(f"invalid model type specified! Available model types: {AVAILABLE_MODELS}")
+
         model_class = MODEL_CLASSES[model_type]
         for param in model_class.REQUIRED_ARGS:
             if param not in model_args.keys():
-                print(f"INFO: model parameter '{param}' not specified -> trying to take from last loaded dataset...")
+                print(f"model parameter '{param}' not specified -> trying to take from last loaded dataset...")
                 if len(self.datasets) < 1:
-                    raise ValueError(f"ERROR: no dataset loaded to take parameter '{param}' from")
+                    raise ValueError(f"no dataset loaded to take parameter '{param}' from")
                 param_val = self.datasets[-1].config.get(param, None)
                 if param_val is None:
-                    raise ValueError(f"ERROR: dataset '{self.datasets[-1].NAME}' doesn't provide parameter '{param}', "
+                    raise ValueError(f"dataset '{self.datasets[-1].NAME}' doesn't provide parameter '{param}', "
                                      f"so it has to be specified on model creation")
                 model_args.update({param: param_val})
         if action_conditional and not model_class.CAN_HANDLE_ACTIONS:
-            print("WARNING: specified model can't handle actions -> argument 'action_conditional' set to False")
+            warnings.warn("specified model can't handle actions -> argument 'action_conditional' set to False")
             action_conditional = False
         model_args.update(action_conditional=action_conditional)
 
@@ -97,31 +157,48 @@ class VPSuite:
         self._model_setup(model)
 
     def _model_setup(self, model, loaded=False):
+        r"""
+
+        Args:
+            model ():
+            loaded ():
+
+        Returns:
+
+        """
         ac_str = "(action-conditional)" if model.config["action_conditional"] else ""
         loaded_str = "loaded" if loaded else "created new"
-        print(f"INFO: {loaded_str} model '{model.NAME}' {ac_str}")
+        print(f"{loaded_str} model '{model.NAME}' {ac_str}")
         total_params = sum(p.numel() for p in model.parameters())
         trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
         print(f" - Model parameters (total / trainable): {total_params} / {trainable_params}")
         self.models.append(model)
 
-    def _get_run_config(self, split="train", **run_args):
-        """ TODO """
-        assert len(self.models) > 0, "No model available. " \
-                                "Load a pretrained model or create a new instance before starting training or test runs"
-        if split == "train":
-            assert len(self.training_sets) > 0, "No training sets loaded. " \
-                                                "Load a dataset in training mode before starting training or test runs"
-        else:
-            assert len(self.test_sets) > 0, "No test sets loaded. " \
-                                                "Load a dataset in test mode before starting training or test runs"
+    def _prepare_run(self, split="train", **run_args):
+        r"""
 
-        with open(self.DEFAULT_RUN_CONFIG, 'r') as tc_file:
+        Args:
+            split ():
+            **run_args ():
+
+        Returns:
+
+        """
+        if len(self.models) == 0:
+            raise RuntimeError("No model available. Load a pretrained model "
+                               "or create a new instance before starting training or test runs")
+        if split == "train" and len(self.training_sets) == 0:
+            raise ValueError("No training sets loaded. Load a dataset in training mode "
+                             "before starting training or test runs")
+        elif split == "test" and len(self.test_sets) == 0:
+            raise ValueError("No test sets loaded. Load a dataset in test mode "
+                             "before starting training or test runs")
+        with open(str(self._DEFAULT_RUN_CONFIG.resolve()), 'r') as tc_file:
             run_config = json.load(tc_file)
 
         # update config
-        assert all([run_arg in run_config.keys() for run_arg in run_args.keys()]), \
-            f"Only the following run arguments are supported: {run_config.keys()}"
+        if not all([run_arg in run_config.keys() for run_arg in run_args.keys()]):
+            raise ValueError(f"Only the following run arguments are supported: {run_config.keys()}")
         run_config.update(run_args)
 
         # seed
@@ -137,19 +214,26 @@ class VPSuite:
 
 # ===== TRAINING ================================================================
 
-    def _prepare_training(self, **training_kwargs):
-        """
+    def _prepare_training(self, dataset_idx, model_idx, **training_kwargs):
+        r"""
         Prepares a single dataset and a single model for training.
+        Args:
+            **training_kwargs ():
+
+        Returns:
+
         """
-        run_config = self._get_run_config("train", **training_kwargs)
+        run_config = self._prepare_run("train", **training_kwargs)
+
+        try:
+            dataset : DatasetWrapper = self.training_sets[dataset_idx]
+            model : VideoPredictionModel = self.models[model_idx]
+        except IndexError:
+            raise ValueError("given indices for model and/or dataset are invalid")
 
         # prepare dataset
-        dataset : DatasetWrapper = self.training_sets[-1]
         dataset.set_seq_len(run_config["context_frames"], run_config["pred_frames"], run_config["seq_step"])
-        assert dataset.is_ready, "ERROR TODO"
-
-        # prepare model
-        model = self.models[-1]
+        assert dataset.is_ready, "dataset is not ready even though set_seq_len has just been called"
 
         # compat checks: run <--> model; model <--> dataset
         check_run_and_model_compat(model, run_config)
@@ -157,10 +241,18 @@ class VPSuite:
 
         return model, dataset, run_config
 
-    def train(self, trial=None, **training_kwargs):
+    def train(self, trial=None, dataset_idx=-1, model_idx=-1, **training_kwargs):
+        r"""
 
+        Args:
+            trial ():
+            **training_kwargs ():
+
+        Returns:
+
+        """
         # PREPARATION
-        model, dataset, run_config = self._prepare_training(**training_kwargs)
+        model, dataset, run_config = self._prepare_training(dataset_idx, model_idx, **training_kwargs)
         train_data, val_data = dataset.train_data, dataset.val_data
         train_loader = DataLoader(train_data, batch_size=run_config["batch_size"], shuffle=True, num_workers=4,
                                   drop_last=True)
@@ -180,8 +272,8 @@ class VPSuite:
             for param, p_dict in optuna_config.items():
                 if "choices" in p_dict.keys():
                     if param == "model_type":
-                        print(f"WARNING: hyperopt across model and dataset parameters is not yet supported "
-                              f"-> using {model.NAME}")
+                        warnings.warn(f"hyperopt across model and dataset parameters is not yet supported "
+                                      f"-> using {model.NAME}")
                     run_config[param] = trial.suggest_categorical(param, p_dict["choices"])
                 else:
                     suggest = trial.suggest_int if p_dict["type"] == "int" else trial.suggest_float
@@ -216,7 +308,7 @@ class VPSuite:
         # LOSSES AND MEASUREMENT
         loss_provider = PredictionLossProvider(config)
         if config['val_rec_criterion'] not in config['losses_and_scales']:
-            raise ValueError(f"ERROR: Validation criterion '{config['val_rec_criterion']}' has to be "
+            raise ValueError(f"Validation criterion '{config['val_rec_criterion']}' has to be "
                              f"one of the chosen losses: {list(config['losses_and_scales'].keys())}")
         if config["opt_direction"] == "maximize":
             def loss_improved(cur_loss, best_loss): return cur_loss > best_loss
@@ -255,7 +347,7 @@ class VPSuite:
                 vis_out_path = out_path / f"vis_ep_{epoch+1:03d}"
                 vis_out_path.mkdir()
                 visualize_vid(val_data, config["context_frames"], config["pred_frames"], model,
-                              config["device"], vis_out_path, dataset.img_processor, num_vis=10)
+                              config["device"], vis_out_path, num_vis=10)
 
                 if not config["no_wandb"]:
                     vid_filenames = sorted(os.listdir(str(vis_out_path)))
@@ -273,18 +365,27 @@ class VPSuite:
         wandb.finish()
         return best_val_loss  # return best validation loss for hyperparameter optimization
 
-    def hyperopt(self, optuna_config=None, n_trials=30, **run_kwargs):
+    def hyperopt(self, optuna_config, n_trials=30, dataset_idx=-1, model_idx=-1, **run_kwargs):
+        r"""
 
+        Args:
+            optuna_config ():
+            n_trials ():
+            **run_kwargs ():
+
+        Returns:
+
+        """
         from functools import partial
-        from vp_suite.utils.utils import _check_optuna_config
-        run_config = self._get_run_config(**run_kwargs)
-        _check_optuna_config(optuna_config)
+        from vp_suite.utils.utils import check_optuna_config
+        run_config = self._prepare_run(**run_kwargs)
+        check_optuna_config(optuna_config)
         run_config["optuna"] = optuna_config
         try:
             import optuna
         except ImportError:
             raise ImportError("Importing optuna failed -> install it or use the code without the 'use-optuna' flag.")
-        optuna_program = partial(self.train, **run_kwargs)
+        optuna_program = partial(self.train, dataset_idx=dataset_idx, model_idx=model_idx, **run_kwargs)
         study = optuna.create_study(direction=run_config["opt_direction"])
         study.optimize(optuna_program, n_trials=n_trials)
         print("\nHyperparameter optimization complete. Best performing parameters:")
@@ -294,16 +395,21 @@ class VPSuite:
 # ===== TESTING ================================================================
 
     def _prepare_testing(self, **run_kwargs):
-        """
+        r"""
         Prepares multiple datasets and models for testing
+        Args:
+            **run_kwargs ():
+
+        Returns:
+
         """
-        run_config = self._get_run_config("test", **run_kwargs)
+        run_config = self._prepare_run("test", **run_kwargs)
 
         # prepare datasets for testing
         test_sets : List[DatasetWrapper] = self.test_sets
         for test_set in test_sets:
             test_set.set_seq_len(run_config["context_frames"], run_config["pred_frames"], run_config["seq_step"])
-            assert test_set.is_ready, "ERROR TODO"
+            assert test_set.is_ready, "test set is not ready even though set_seq_len has just been called"
 
         # compat-check run arguments against models and skip incompatible models
         test_models = []
@@ -336,19 +442,30 @@ class VPSuite:
         return test_sets_and_model_lists, run_config
 
     def _test_on_dataset(self, model_info_list, dataset, run_config, brief_test):
+        r"""
 
+        Args:
+            model_info_list ():
+            dataset ():
+            run_config ():
+            brief_test ():
+
+        Returns:
+
+        """
         # PREPARATION
         test_data = dataset.test_data
         test_loader = DataLoader(test_data, batch_size=1, shuffle=True, num_workers=0)
-        iter_loader = iter(test_loader)
+        if len(test_loader) < 1:
+            raise RuntimeError("loaded dataset does not contain any data (len < 1)")
         test_mode = "brief" if brief_test else "full"
-        eval_length = 10 if test_mode == "brief" else len(test_loader)
-        assert eval_length > 0, "ERROR: length of test loader < 1 -> no evaluation possible"
+        eval_length = min(len(test_loader), 10) if test_mode == "brief" else len(test_loader)
 
         # assemble and save combined configuration
         config = {**run_config, **dataset.config, "device": self.device}
 
         # evaluation / metric calc.
+        iter_loader = iter(test_loader)
         with torch.no_grad():
             metric_provider = PredictionMetricProvider(config)
 
@@ -380,13 +497,12 @@ class VPSuite:
                 vis_out_dir = Path(model.model_dir) / f"vis_{timestamp('test')}"
                 vis_out_dir.mkdir()
                 visualize_vid(test_data, config["context_frames"], config["pred_frames"],
-                              model, config["device"], vis_out_dir, dataset.img_processor, vis_idx=vis_idx)
+                              model, config["device"], vis_out_dir, vis_idx=vis_idx)
 
         # log or display metrics
         if eval_length > 0:
             wandb_full_suffix = f"{test_mode} test"
             for i, (model, _, _, model_metrics_per_dp) in enumerate(model_info_list):
-
                 # model_metrics_per_dp: list of N lists of F metric dicts (called D).
                 # each D_{n, f} contains all calculated metrics for datapoint 'n' and a prediction horizon of 'f' frames.
                 # -> Aggregate these metrics over all n, keeping the specific metrics/prediction horizons separate
@@ -421,6 +537,15 @@ class VPSuite:
                             print(f" -> {k}: {v}")
 
     def test(self, brief_test=False, **run_kwargs):
+        r"""
+
+        Args:
+            brief_test ():
+            **run_kwargs ():
+
+        Returns:
+
+        """
         test_sets_and_model_lists, run_config = self._prepare_testing(**run_kwargs)
         for test_set, model_info_list in test_sets_and_model_lists:
             self._test_on_dataset(model_info_list, test_set, run_config, brief_test)
